@@ -8,11 +8,13 @@ package Net::Prometheus::Metric;
 use strict;
 use warnings;
 
-our $VERSION = '0.01';
+our $VERSION = '0.02';
 
 use Carp;
 our @CARP_NOT = qw( Net::Prometheus );
 
+use Net::Prometheus::Types qw( Sample MetricSamples );
+
 =head1 NAME
 
 C<Net::Prometheus::Metric> - the base class for observed metrics
@@ -105,15 +107,30 @@ sub new
    my $class = shift;
    my %args = @_;
 
-   my $fullname = join "_", grep { defined } $args{namespace}, $args{subsystem}, $args{name};
-
+   defined $args{name} or
+      croak "Required 'name' argument missing";
    defined $args{help} or
       croak "Required 'help' argument missing";
 
+   my $fullname = join "_", grep { defined } $args{namespace}, $args{subsystem}, $args{name};
+
+   my $labellist = $args{labels} || [];
+
+   # See
+   #   https://prometheus.io/docs/concepts/data_model/#metric-names-and-labels
+   $fullname =~ m/^[a-zA-Z_:][a-zA-Z0-9_:]*$/ or
+      croak "Invalid metric name '$fullname'";
+
+   $_ =~ m/^[a-zA-Z_][a-zA-Z0-9_]*$/ or
+      croak "Invalid label name '$_'" for @$labellist;
+   $_ =~ m/^__/ and
+      croak "Label name '$_' is reserved" for @$labellist;
+
    return bless {
       fullname => $fullname,
       help => $args{help},
-      labels => $args{labels} // [],
+      labels => $labellist,
+      labelvalues => {},
    }, $class;
 }
 
@@ -174,53 +191,78 @@ sub labels
    @values == $labelcount or
       croak "Too many values given for labels";
 
-   # Encode \x00 or \x01 as \x{01}0 or \x{01}1 in order to escape the \x00 but
-   # preserve full leixcal ordering
-   $_ =~ s/\x01/\x011/g, $_ =~ s/\x00/\x010/g for @values;
+   length $values[$_] or
+      croak "Value for $self->{labels}[$_] may not empty" for 0 .. $#values;
+
+   my $labelkey = join "\x00", map {
+      # Encode \x00 or \x01 as \x{01}0 or \x{01}1 in order to escape the \x00
+      # but preserve full leixcal ordering
+      my $value = $_;
+      $value =~ s/\x01/\x011/g;
+      $value =~ s/\x00/\x010/g;
+      $value;
+   } @values;
+
+   $self->{labelvalues}{$labelkey} = \@values;
 
    return Net::Prometheus::Metric::_Child->new(
-      $self, join "\x00", @values,
+      $self, $labelkey
    );
 }
 
-=head2 render
+=head2 make_sample
 
-   @strs = $metric->render
+   $sample = $metric->make_sample( $suffix, $labelkey, $value )
 
-Returns a list of string in the Prometheus text exposition format containing
-the current values of the metric, along with the descriptive help text and
-type information. These strings will individually contain their own linefeeds;
-the result can be joined simply with
-
-   $str = join "", $metric->render
+Returns a new L<Net::Prometheus::Types/Sample> structure to represent the
+given value, by expanding the opaque C<$labelkey> value into its actual label
+names and values and appending the given suffix (which may be an empty string)
+to the metric's fullname. If provided, the suffix will be separated by an
+underscore C<'_'>.
 
 =cut
 
-sub _render_value
+sub make_sample
 {
    my $self = shift;
-   my ( $varname, $labelkey, $value ) = @_;
+   my ( $suffix, $labelkey, $value ) = @_;
 
-   return "$varname $value\n" if $labelkey eq "";
+   my $labelnames  = $self->{labels};
+   my $labelvalues = $self->{labelvalues}{$labelkey};
 
-   my $labels = $self->{labels};
-   my @labvalues = split m/\x00/, $labelkey;
-
-   return $varname . "{" . join( ",", map {
-      qq($labels->[$_]="$labvalues[$_]")
-   } 0 .. $#labvalues ) . "} $value\n";
+   return Sample(
+      ( $suffix ? $self->fullname . "_$suffix" : $self->fullname ),
+      [ map { $labelnames->[$_], $labelvalues->[$_] } 0 .. $#$labelnames ],
+      $value,
+   );
 }
 
-sub render
+sub collect
 {
    my $self = shift;
 
-   my $fullname = $self->fullname;
-   my $type = $self->_type;
+   return MetricSamples(
+      $self->fullname, $self->_type, $self->{help},
+      [ $self->samples ],
+   );
+}
 
-   return
-      "# HELP $fullname $self->{help}\n",
-      "# TYPE $fullname $type\n";
+=head2 samples
+
+   @samples = $metric->samples
+
+An abstract method in this class, this method is intended to be overriden by
+subclasses.
+
+Called during the value collection process, this method should return a list
+of L<Net::Prometheus::Types/Sample> instances containing the values to report
+from this metric.
+
+=cut
+
+sub samples
+{
+   croak "Abstract Net::Prometheus::Metric->samples invoked directly";
 }
 
 package
@@ -238,6 +280,9 @@ sub new
    return bless [ $metric, $labelkey ], $class;
 }
 
+sub metric   { shift->[METRIC] }
+sub labelkey { shift->[LABELKEY] }
+
 foreach my $method (qw( set inc dec observe )) {
    my $childmethod = "_${method}_child";
 
