@@ -9,7 +9,10 @@ use strict;
 use warnings;
 use base qw( Net::Prometheus::Metric );
 
-our $VERSION = '0.01';
+our $VERSION = '0.02';
+
+use Carp;
+use List::Util 1.33 qw( any );
 
 use constant _type => "summary";
 
@@ -63,6 +66,10 @@ __PACKAGE__->make_childlabel_method( $_ ) for qw( observe );
 sub new
 {
    my $class = shift;
+   my %opts = @_;
+
+   $opts{labels} and any { $_ eq "quantile" } @{ $opts{labels} } and
+      croak "A Summary may not have a label called 'quantile'";
 
    my $self = $class->SUPER::new( @_ );
 
@@ -91,17 +98,17 @@ sub _observe_child
    $self->{sums}  {$labelkey} += $value;
 }
 
-sub render
+sub samples
 {
    my $self = shift;
 
    my $counts = $self->{counts};
    my $sums   = $self->{sums};
-   my $fullname = $self->fullname;
 
-   return $self->SUPER::render,
-      map { $self->_render_value( "${fullname}_count", $_, $counts->{$_} ),
-            $self->_render_value( "${fullname}_sum",   $_, $sums->{$_} ) } sort keys %$counts;
+   return map {
+      $self->make_sample( count => $_, $counts->{$_} ),
+      $self->make_sample( sum   => $_, $sums->{$_} )
+   } sort keys %$counts;
 }
 
 =head1 AUTHOR
